package de.hamster.scratch;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.image.BufferedImage;
import java.util.ArrayList;

import javax.swing.JOptionPane;
import javax.swing.JPanel;

import de.hamster.editor.view.TextArea;
import de.hamster.scratch.Renderable.Type;
import de.hamster.scratch.gui.ClickHandler;
import de.hamster.scratch.gui.ContextMenu;
import de.hamster.scratch.gui.ContextMenuHandler;
import de.hamster.scratch.gui.DeleteMethodException;
import de.hamster.scratch.gui.ElementList;
import de.hamster.scratch.gui.InvalidIdentifierException;
import de.hamster.scratch.gui.LocationChangeHandler;
import de.hamster.scratch.gui.MenuItem;
import de.hamster.scratch.gui.MethodChangeHandler;
import de.hamster.scratch.gui.NextMethodHandler;
import de.hamster.scratch.gui.OptionsPanel;
import de.hamster.scratch.gui.RefreshHandler;
import de.hamster.scratch.gui.RenameFunctionFrame;
import de.hamster.scratch.gui.RenameMethodException;
import de.hamster.scratch.gui.ScratchScrollBar;
import de.hamster.scratch.gui.TabButton;
import de.hamster.scratch.gui.TabChangedHandler;
import de.hamster.scratch.gui.TabOpenedHandler;
import de.hamster.scratch.gui.TabPanel;

/**
 * Diese Klasse ist die Herzstck GUI des Programms. Diese Klasse
 * ist das Hauptpanel auf dem alles gerendert wird und das die
 * interne Logik zusammenfasst. Das Panel muss an der richtigen
 * Stelle im Hamster-Programm eingebunden werden.
 * @author HackZ
 *
 */
public class ScratchPanel extends JPanel implements MouseListener, MouseMotionListener, TabChangedHandler, RefreshHandler, ContextMenuHandler, MethodChangeHandler {
	/**
	 * 
	 */
	private static final long serialVersionUID = -6049901089566290960L;
	public static Color BACKGROUND_COLOR = new Color(124, 128, 131);
	public static Color SECONDARY_COLOR = new Color(102, 105, 107);
	public static Color THIRD_COLOR = new Color(117, 120, 123);
	public static int OPTIONS_WIDTH = 147;
	public static int OPTIONS_HEIGHT = 94;
	public static int TAB_PANEL_HEIGHT = 40;
	
	private static BufferedImage imgTop;
	private static BufferedImage imgBottom;
	private static BufferedImage imgLeft;
	private static BufferedImage imgRight;
	private static BufferedImage imgTopLeft;
	private static BufferedImage imgTopRight;
	private static BufferedImage imgBottomLeft;
	private static BufferedImage imgBottomRight;

	private Method method;
	private TextArea textArea;
	private Renderable selected;
	private Renderable docking;
	private Renderable rootElement;
	private ArrayList<Renderable> renderables;
	private Point tempClick;
	private ElementList elements;
	private OptionsPanel optionsPanel;
	private TabPanel tabPanel;
	private ScratchScrollBar vScrollbar;
	private ScratchScrollBar hScrollbar;
	private ContextMenu contextMenu;
	
	private boolean panelLocked = false; // dibo 290710
	
	public ScratchPanel() {
		this.textArea = null;
		this.addMouseListener(this);
		this.addMouseMotionListener(this);
		this.setLayout(null);
		this.setBackground(BACKGROUND_COLOR);
		loadImages();

		// Context menu on TOP
		contextMenu = new ContextMenu();
		contextMenu.setBounds(0, 0, getWidth(), getHeight());
		contextMenu.addRefreshHandler(this);
		this.add(contextMenu);
		
		// Elementliste
		elements = new ElementList(this);
		elements.setBounds(0, OPTIONS_HEIGHT, OPTIONS_WIDTH, this.getHeight() - OPTIONS_HEIGHT);
		this.add(elements);
		
		// Optionsmenue
		optionsPanel = new OptionsPanel(elements, this);
		optionsPanel.setBounds(0, 0, OPTIONS_WIDTH, OPTIONS_HEIGHT);
		this.add(optionsPanel);
		
		// Scrollbars
		vScrollbar = new ScratchScrollBar(ScratchScrollBar.VERTICAL);
		vScrollbar.setBounds(getWidth() - 25, 10, ScratchScrollBar.WIDTH, getHeight() - 20);
		vScrollbar.addLocationChangeHandler(new VerticalLocationHandler());
		this.add(vScrollbar);
		
		hScrollbar = new ScratchScrollBar(ScratchScrollBar.HORIZONTAL);
		hScrollbar.setBounds(OPTIONS_WIDTH + 10, getHeight() - 25, getWidth() - 20 - OPTIONS_WIDTH, ScratchScrollBar.WIDTH);
		hScrollbar.addLocationChangeHandler(new HorizontalLocationHandler());
		this.add(hScrollbar);
		
		// TabPanel
		tabPanel = new TabPanel();
		tabPanel.setBounds(OPTIONS_WIDTH, 0, getWidth() - OPTIONS_WIDTH, TAB_PANEL_HEIGHT);
		tabPanel.addTabChangedHandler(this);
		tabPanel.addContextMenuHandler(this);
		tabPanel.addMethodChangeHandler(this);
		tabPanel.addTabOpenedHandler(new OpenedMethodHandler());
		this.add(tabPanel);
	}
	
	/**
	 * Beim jeden Erstellen des Panels wird versucht die statischen
	 * Hintergrundbilder zu laden. Sollte bereits eines der Bilder
	 * geladen worden sein, so wird die Methode abgebrochen.
	 */
	private void loadImages() {
		if (imgTop != null)
			return; // Bilder sind bereits geladen
		
		imgTop = ScratchUtils.getImage("scratch/backgrounds/top_elementlist.png");
		imgBottom = ScratchUtils.getImage("scratch/backgrounds/bottom_elementlist.png");
		imgLeft = ScratchUtils.getImage("scratch/backgrounds/left.png");
		imgRight = ScratchUtils.getImage("scratch/backgrounds/right.png");
		imgTopLeft = ScratchUtils.getImage("scratch/backgrounds/top_left_elementlist.png");
		imgTopRight = ScratchUtils.getImage("scratch/backgrounds/top_right_elementlist.png");
		imgBottomLeft = ScratchUtils.getImage("scratch/backgrounds/bottom_left_elementlist.png");
		imgBottomRight = ScratchUtils.getImage("scratch/backgrounds/bottom_right_elementlist.png");
	}

	/**
	 * Setzt die zugeordnete TextArea (nicht verndern)
	 */
	public void setTextArea(TextArea area) {
		this.textArea = area;
		((ScratchHamsterFile)(textArea.getFile())).getProgram().addRefreshHandler(this);
		((ScratchHamsterFile)(textArea.getFile())).getProgram().addNextMethodHandler(new OnNextMethodHandler());
		
		ArrayList<String> openedTabs = getStorageController().getOpenedMethods();
		for (String s : openedTabs)
			tabPanel.addTab(s);
		
		optionsPanel.loadLastActive();
	}

	/**
	 * Liefert die zugeordnete TextArea (nicht verndern)
	 */
	public TextArea getTextArea() {
		return textArea;
	}
	
	/**
	 * Liefert den StorageController, in dem die Methoden gespeichert sind.
	 * Dieser kann aus dem TextArea ausgelesen werden.
	 * @return
	 */
	public StorageController getStorageController() {
		return ((ScratchHamsterFile)(textArea.getFile())).getProgram().getProgram();
	}

	/**
	 * Wird aufgerufen, wenn ein Scratch-Programm gestartet bzw. beendet wird.
	 * Whrend einer Programm-Ausfhrung sollte das Programm nicht verndert
	 * werden knnen (locked == true)
	 */
	public void setLocked(boolean locked) {
		// begin dibo 290710
		if (locked == this.panelLocked) {
			return;
		}
		this.panelLocked = locked;
		// end dibo 290710
		
		elements.setLocked(locked);
		if (locked) {
			this.removeMouseListener(this);
		} else {
			this.addMouseListener(this);
		}
	}
	
	/**
	 * Da beim Docken mehrere Parameter zurckgeliefert werden mssten, wird
	 * diese Methode verwendet, um whrend des dockVorgang berechnet, an
	 * welches Objekt angedockt werden kann.
	 * @param r
	 */
	public void setDocking(Renderable r) {
		this.docking = r;
	}
	
	/**
	 * Teil der Methode paint(Graphics g). Weitere
	 * Elemente werden gerendert.
	 * @param g
	 */
	public void render(Graphics g) {
		rootElement.render(g);
		for (Renderable r : renderables) {
			if (r == selected)
				continue;
			r.render(g);
		}
		
		highlightDockings(g);
		paintBorders(g);
		paintChildren(g);
		
		if (selected != null) {
			selected.renderShadow(g);
			selected.render(g);
		}
	}
	
	/**
	 * Zeichnet die Randbilder. Das Look-and-Feel von Scratch.
	 * @param g
	 */
	private void paintBorders(Graphics g) {
//		g.drawImage(imgTop, 0, 0, getWidth(), 8, BACKGROUND_COLOR, null);
		for (int i = 0; i < imgTop.getHeight(); i++) {
			g.setColor(new Color(imgTop.getRGB(0, i)));
			g.drawLine(OPTIONS_WIDTH, i + TAB_PANEL_HEIGHT, getWidth(), i + TAB_PANEL_HEIGHT);
		}
		
//		g.drawImage(imgBottom, 0, getHeight() - 8, getWidth(), 8, BACKGROUND_COLOR, null);
		for (int i = 0; i < imgBottom.getHeight(); i++) {
			int top = i + getHeight() - imgBottom.getHeight();
			g.setColor(new Color(imgBottom.getRGB(0, i)));
			g.drawLine(OPTIONS_WIDTH, top, getWidth(), top);
		}
		
//		g.drawImage(imgLeft, 0, 0, 8, getHeight(), BACKGROUND_COLOR, null);
		for (int i = 0; i < imgLeft.getWidth(); i++) {
			g.setColor(new Color(imgLeft.getRGB(i, 0)));
			g.drawLine(OPTIONS_WIDTH + i, TAB_PANEL_HEIGHT, OPTIONS_WIDTH + i, getHeight());
		}
		
//		g.drawImage(imgRight, getWidth() - 5, 0, 5, getHeight(), BACKGROUND_COLOR, null);
		for (int i = 0; i < imgRight.getWidth(); i++) {
			int left = i + getWidth() - imgRight.getWidth();
			g.setColor(new Color(imgRight.getRGB(i, 0)));
			g.drawLine(left, TAB_PANEL_HEIGHT, left, getHeight());
		}
		
		g.drawImage(imgTopLeft, OPTIONS_WIDTH, TAB_PANEL_HEIGHT, null);
		g.drawImage(imgTopRight, getWidth() - imgTopRight.getWidth(), TAB_PANEL_HEIGHT, null);
		g.drawImage(imgBottomLeft, OPTIONS_WIDTH, getHeight() - imgBottomLeft.getHeight(), null);
		g.drawImage(imgBottomRight, getWidth() - imgBottomRight.getWidth(), getHeight() - imgBottomRight.getHeight(), null);
	}
	
	@Override
	public void setBounds(int x, int y, int width, int height) {
		super.setBounds(x, y, width, height);
		updateScrollbars();
		elements.setBounds(0, OPTIONS_HEIGHT, OPTIONS_WIDTH, this.getHeight() - OPTIONS_HEIGHT);
		vScrollbar.setBounds(getWidth() - 25, 10 + TAB_PANEL_HEIGHT, ScratchScrollBar.WIDTH, getHeight() - TAB_PANEL_HEIGHT - (hScrollbar.isVisible() ? 40 : 20));
		hScrollbar.setBounds(OPTIONS_WIDTH + 10, getHeight() - 25, getWidth() - (vScrollbar.isVisible() ? 40 : 20) - OPTIONS_WIDTH, ScratchScrollBar.WIDTH);
		tabPanel.setBounds(OPTIONS_WIDTH, 0, getWidth() - OPTIONS_WIDTH, TAB_PANEL_HEIGHT);
		contextMenu.setBounds(0, 0, getWidth(), getHeight());
		contextMenu.setVisible(false);
	}
	
	@Override
	public void setBounds(Rectangle r) {
		setBounds(r.x, r.y, r.width, r.height);
	}
	
	/**
	 * Rendert neben den Elementen die DockingBoxen, an die
	 * das selektierte Element angedockt werden kann.
	 * @param g
	 */
	private void highlightDockings(Graphics g) {
		if (selected == null)
			return;
		
		// Andocken an Root Element
		int dockIndex = rootElement.dockingTest(selected, this);
		if (dockIndex >= 0) {
			docking.highlightDocking(g, dockIndex);
			return;
		} else if (dockIndex == -2) {
			docking.highlightParentDocking(g);
			return;
		} else if (dockIndex == -3) {
			docking.highlightNextDocking(g);
			return;
		}
		
		for (Renderable r : renderables) {
			dockIndex = r.dockingTest(selected, this);
			if (dockIndex >= 0) {
				docking.highlightDocking(g, dockIndex);
				return;
			} else if (dockIndex == -2) {
				docking.highlightParentDocking(g);
				return;
			} else if (dockIndex == -3) {
				docking.highlightNextDocking(g);
				return;
			}
		}
	}
	
	/**
	 * Fgt das bergebene Renderable hinzu, so dass dieses
	 * im HauptFenster angezeigt wird.
	 * @param r
	 */
	public void addRenderable(Renderable r) {
		renderables.add(r);
		modified();
		repaint();
	}
	
	/**
	 * Fgt ein Renderable an Position x, y in der Elementenliste
	 * hinzu und fhrt gleichzeitig ein mouseDown Event aus
	 * @param r
	 * Das Renderable, das hinzugefgt werden soll
	 * @param x
	 * Koordinate x
	 * @param y
	 * Koordinate y
	 */
	public void addRenderable(Renderable r, MouseEvent e) {
		addRenderable(r);
		selected = r;
		tempClick = new Point(e.getX(), e.getY());
		modified();
	}
	
	@Override
	public void paint(Graphics g) {
		paintBackground(g);
		synchronized (this) {
			render(g);
		}
	}
	
	/**
	 * Zeichnet den Hintergrund (Gestreift). Look-and-Feel von Scratch.
	 * @param g
	 */
	private void paintBackground(Graphics g) {
		g.setColor(BACKGROUND_COLOR);
		g.fillRect(OPTIONS_WIDTH, 0, getWidth() - OPTIONS_WIDTH, getHeight());
		
		int i = OPTIONS_WIDTH;
		g.setColor(SECONDARY_COLOR);
		while (i <= getWidth()) {
			g.drawLine(i, 0, i, getHeight());
			i += 4;
		}
		
		i = OPTIONS_WIDTH + 2;
		g.setColor(THIRD_COLOR);
		while (i <= getWidth()) {
			g.drawLine(i, 0, i, getHeight());
			i += 4;
		}
	}
	
	/**
	 * Liefert die Gesamthhe aller Elemente. Notwendig, um die ScrollBars
	 * so anzupassen, dass diese bis zu der maximalen Hhe gescrollt werden
	 * knnen.
	 * @return
	 */
	public int getInnerHeight() {
		int h = 0;
		for (Renderable r : renderables) {
			BoundingBox temp = r.getGlobalBounding();
			int tempH = temp.getY() + temp.getHeight();
			if (tempH > h)
				h = tempH;
		}
		
		BoundingBox temp = rootElement.getGlobalBounding();
		int tempH = temp.getY() + temp.getHeight();
		if (tempH > h)
			h = tempH;
			
		return h + (hScrollbar.isVisible() ? 25 : 0) + 20 + method.getTop();
	}
	
	/**
	 * Liefert die Gesamtbreite aller Elemente. Notwendig, um die ScrollBars
	 * so anzupassen, dass diese bis zu der maximalen Breite gescrollt werden
	 * knnen.
	 * @return
	 */
	public int getInnerWidth() {
		int w = 0;
		for (Renderable r : renderables) {
			BoundingBox temp = r.getGlobalBounding();
			int tempW = temp.getX() + temp.getWidth();
			if (tempW > w)
				w = tempW;
		}
		
		BoundingBox temp = rootElement.getGlobalBounding();
		int tempW = temp.getX() + temp.getWidth();
		if (tempW > w)
			w = tempW;
			
		return w + (vScrollbar.isVisible() ? 25 : 0) + 20 + method.getLeft();
	}
	
	/**
	 * Berechnet die notwendige Breite und Hhe fr die
	 * Darstellung aller Elemente und zeigt die Scroll-
	 * balken, falls notwendig.
	 */
	private void updateScrollbars() {
		updateScrollbarsIntern();
		updateScrollbarsIntern();
		vScrollbar.setBounds(getWidth() - 25, 10 + TAB_PANEL_HEIGHT, ScratchScrollBar.WIDTH, getHeight() - TAB_PANEL_HEIGHT - (hScrollbar.isVisible() ? 40 : 20));
		hScrollbar.setBounds(OPTIONS_WIDTH + 10, getHeight() - 25, getWidth() - (vScrollbar.isVisible() ? 40 : 20) - OPTIONS_WIDTH, ScratchScrollBar.WIDTH);
	}
	
	/**
	 * Passt die Hhen und Breiten der ScrollBars an.
	 */
	private void updateScrollbarsIntern() {
		if (getHeight() == 0)
			return;
		
		int vScrollHeight = getInnerHeight() - getHeight();
		if (vScrollHeight > 0) {
			vScrollbar.setVisible(true);
			vScrollbar.setMaximum(vScrollHeight);
		}
		else {
			vScrollbar.setVisible(false);
			for (Renderable r : renderables)
				r.moveRelative(0, method.getTop());
			rootElement.moveRelative(0, method.getTop());
			method.setTop(0);
		}
		
		int hScrollWidth = getInnerWidth() - getWidth();
		if (hScrollWidth > 0) {
			hScrollbar.setVisible(true);
			hScrollbar.setMaximum(hScrollWidth);
		}
		else {
			hScrollbar.setVisible(false);
			for (Renderable r : renderables)
				r.moveRelative(method.getLeft(), 0);
			rootElement.moveRelative(method.getLeft(), 0);
			method.setLeft(0);
		}
	}
	
	/**
	 * Prft, ob der bergebene JavaIdentifier fr eine neue Methode
	 * mglich ist.
	 * @param name
	 * Name fr eine neue Methode.
	 * @throws InvalidIdentifierException
	 * wird geworfen, falls
	 * <ol>
	 * <li>Der Bezeichner hat keine gltige Javakonvention.</li>
	 * <li>Der Bezeichner ist bereits vergeben.</li>
	 * <li>Der Bezeichner ist ein reserviertes Javakeyword.</li>
	 * </ol>
	 */
	public void checkJavaIdentifier(String name) throws InvalidIdentifierException {
		getStorageController().allocName(name);
	}
	
	public void createMethod(String name, Type type) {
		switch (type) {
		case VOID:
			getStorageController().addVoidMethod(name);
			break;
		case BOOLEAN:
			getStorageController().addBooleanMethod(name);
			break;
		}
		tabPanel.addTab(name);
		optionsPanel.refreshElementList();
		modified();
	}
	
	/**
	 * Setzt fest, dass das Panel modifiziert wurde, so dass der
	 * Speicherbutton enabled wird.
	 */
	public void modified() {
		if (textArea == null)
			return;
		
		textArea.getFile().setModified(true);
	}
	
	/**
	 * Setzt fest, dass das Panel nicht modifiziert wurde, so dass der
	 * Speicherbutton disabled wird.
	 */
	public void unmodified() {
		if (textArea == null)
			return;
		
		textArea.getFile().setModified(false);
	}
	
	/**
	 * ffnet einen Tab mit dem bergebenen Namen.
	 * @param name
	 */
	public void openTab(String name) {
		if (getStorageController().existsMethod(name))
			tabPanel.addTab(name);
	}
	
	@Override
	public void openContextMenu(ArrayList<MenuItem> menuItems, int x, int y) {
		contextMenu.openMenu(menuItems, x, y);
	}
	
	/**
	 * ffnet ein ContextMenu an der bergebenen Position
	 * @param x
	 * x-Koordinate des Menus.
	 * @param y
	 * y-Koordinate des Menus.
	 */
	public void openContextMenu(int x, int y) {
		for (int i = 0; i < renderables.size(); i++) {
			Renderable temp = renderables.get(i).hitTest(x, y);
			if (temp != null) {
				openContextMenu(temp, x, y, false);
				return;
			}
		}
		
		Renderable temp = rootElement.hitTest(x, y);
		if (temp != null)
			openContextMenu(temp, x, y, false);
	}
	
	/**
	 * ffnet ein ContextMenu an der bergebenen Position, das das
	 * bergebene Renderable bearbeiten soll.
	 * @param elem
	 * Das Element, das mit diesem Menu bearbeitet wird.
	 * @param x
	 * x-Koordinate des Menus.
	 * @param y
	 * y-Koordinate des Menus.
	 * @param elementList
	 * Wurde diese Methode aus der ElementList heraus aufgerufen.
	 */
	public void openContextMenu(Renderable elem, int x, int y, boolean elementList) {
		ArrayList<MenuItem> items = new ArrayList<MenuItem>();
		
		boolean en = getStorageController().existsMethod(elem.getName());
		
		MenuItem temp = new MenuItem("ffnen", new OpenClickHandler(elem.getName()));
		temp.setEnabled(en);
		items.add(temp);
		
		MenuItem renameItem = new MenuItem("Umbenennen", new RenameClickHandler(elem.getName()));
		renameItem.setEnabled(en);
		items.add(renameItem);
		
		String delName = (elementList) ? "Lschen" : "Entfernen";
		ClickHandler delHandler = (elementList) ? new DeleteClickHandler(elem.getName()) : new RemoveClickHandler(elem);
		MenuItem deleteItem = new MenuItem(delName, delHandler);
		deleteItem.setEnabled(!elementList || en);
		items.add(deleteItem);
		
		openContextMenu(items, x, y);
 	}

	@Override
	public void mouseClicked(MouseEvent e) {
		if (e.getButton() == MouseEvent.BUTTON3) {
			openContextMenu(e.getX(), e.getY());
			return;
		}
		
		if (e.getButton() != MouseEvent.BUTTON1 || (e.getClickCount() % 2) != 0)
			return;
		
		for (int i = 0; i < renderables.size(); i++) {
			Renderable temp = renderables.get(i).hitTest(e.getX(), e.getY());
			if (temp != null) {
				openTab(temp.getName());
				return;
			}
		}
		
		Renderable temp = rootElement.hitTest(e.getX(), e.getY());
		if (temp != null)
			openTab(temp.getName());
	}

	@Override
	public void mouseEntered(MouseEvent arg0) {	}

	@Override
	public void mouseExited(MouseEvent arg0) { }

	@Override
	public void mousePressed(MouseEvent e) {
		//textArea.getFile().setModified(true);
		
		if (e.getButton() != MouseEvent.BUTTON1)
			return;
		
		for (int i = renderables.size() - 1; i >= 0; i--) {
			Renderable temp = renderables.get(i).hitTest(e.getX(), e.getY());
			if (temp != null) {
				selected = temp;
				Renderable selectedParent = selected.parent;
				
				if (selected.removeFromParent()) {
					addRenderable(selected);
					selectedParent.updateBounds();
					selectedParent.getRootElement().updateBounds();
					selectedParent.getRootElement().updateChilds();
				}
				else {
					// Bring to top
					renderables.remove(selected);
					renderables.add(selected);
				}
				tempClick = e.getPoint();
				return;
			}
		}
		
		// Abdocken vom Root Element
		Renderable temp = rootElement.hitTest(e.getX(), e.getY());
		if (temp != null) {
			selected = temp;
			Renderable selectedParent = selected.parent;
			
			if (selected.removeFromParent()) {
				addRenderable(selected);
				selectedParent.updateBounds();
				selectedParent.getRootElement().updateBounds();
				selectedParent.getRootElement().updateChilds();
			}
			else {
				// Bring to top
				renderables.remove(selected);
				renderables.add(selected);
			}
			tempClick = e.getPoint();
			return;
		}
		
		selected = null;
	}

	@Override
	public void mouseReleased(MouseEvent e) {
		if (e.getButton() != MouseEvent.BUTTON1)
			return;
		
		if (selected == null)
			return;
		
		// Andocken an Root Element
		boolean docked = false;
		int dockIndex = rootElement.dockingTest(selected, this);
		if (dockIndex >= 0) {
			docking.add(selected, dockIndex);
			renderables.remove(selected);
			docking.updateBounds();
			docking.getRootElement().updateChilds();
			docked = true;
		} else if (dockIndex == -2) {
			docking.addAsParent(selected);
			docking.updateBounds();
			selected.getRootElement().updateChilds();
			renderables.remove(docking);
			docked = true;
		} else if (dockIndex == -3) {
			docking.addAsNext(selected);
			selected.updateBounds();
			selected.getRootElement().updateChilds();
			renderables.remove(selected);
			docked = true;
		}
		
		if (!docked) {
			// Andocken an eines der anderen Elemente
			for (Renderable r : renderables) {
				dockIndex = r.dockingTest(selected, this);
				if (dockIndex >= 0) {
					docking.add(selected, dockIndex);
					renderables.remove(selected);
					docking.updateBounds();
					docking.getRootElement().updateChilds();
					break;
				} else if (dockIndex == -2) {
					docking.addAsParent(selected);
					docking.updateBounds();
					selected.getRootElement().updateChilds();
					renderables.remove(docking);
					break;
				} else if (dockIndex == -3) {
					docking.addAsNext(selected);
					selected.updateBounds();
					selected.getRootElement().updateChilds();
					renderables.remove(selected);
					break;
				}
			}
		}
		
		if (selected.getY() < TAB_PANEL_HEIGHT)
			selected.moveRelative(0, TAB_PANEL_HEIGHT - selected.getY());
		
		if (selected.getX() < OPTIONS_WIDTH)
			if (e.getX() < OPTIONS_WIDTH)
				renderables.remove(selected);
			else
				selected.moveTo(OPTIONS_WIDTH, selected.getY());
		
		selected = null;
		updateScrollbars();
		repaint();
	}

	@Override
	public void mouseDragged(MouseEvent e) {
		if (selected == null)
			return;
		
		selected.moveRelative(e.getX() - (int)tempClick.getX(), e.getY() - (int)tempClick.getY());
		modified(); // dibo 290710
		
		tempClick = e.getPoint();
		repaint();
	}

	@Override
	public void mouseMoved(MouseEvent arg0) { }

	@Override
	public void onDelete(TabButton button) {
		onDelete(button.getName());
	}
	
	/**
	 * Lscht die Methode mit dem bergebenen Namen.
	 * @param name
	 * Name der Methode, die gelscht werden soll.
	 */
	private void onDelete(String name) {
		int confirmResult = JOptionPane.showConfirmDialog(null, "Soll die Methode \"" + name + "\" gelscht werden? Dieser Vorgang kann nicht rckgngig gemacht werden!", "Lschen", JOptionPane.YES_NO_OPTION);
		
		if (confirmResult == JOptionPane.YES_OPTION) {
			try {
				getStorageController().deleteMethod(name);
				tabPanel.closeTab(name);
				optionsPanel.refreshElementList();
			} catch (DeleteMethodException ex) {
				JOptionPane.showMessageDialog(null, ex.getMessage(), "Fehler", JOptionPane.OK_OPTION);
			}
		}
	}

	@Override
	public void onRename(String name) {
		new RenameFunctionFrame(this, name);
		modified();
	}
	
	/**
	 * Entfernt das angegebene Element und vereinigt das Vaterelement
	 * mit dem Kindelement.
	 * @param elem
	 * Element, das entfernt werden soll.
	 */
	public void onRemove(Renderable elem) {
		Renderable root = elem.getRootElement();
		Renderable parent = elem.parent;
		Renderable child = elem.next;
		modified();
		
		elem.removeFromParent();
		if (child != null)
			child.removeFromParent();
		
		if (parent == null) {
			if (child != null)
				this.addRenderable(child);
			
			renderables.remove(elem);
			return;
		}
		
		if (child != null)
			parent.addAsNext(child);
		
		renderables.remove(elem);
		parent.updateBounds();
		root.updateBounds();
		root.updateChilds();
		repaint();
	}

	/**
	 * Benennt dalle Elemente mit dem Namen fromName in toName
	 * @param fromName
	 * @param toName
	 */
	public void renameMethod(String fromName, String toName) {
		try {
			getStorageController().rename(fromName, toName);
			tabPanel.renameTab(fromName, toName);
			optionsPanel.refreshElementList();
			modified();
			repaint();
		} catch (RenameMethodException ex) {
			JOptionPane.showMessageDialog(null, ex.getMessage(), "Fehler", JOptionPane.OK_OPTION);
		}
	}

	@Override
	public void tabChanged(TabButton button) {
		Method method = getStorageController().getMethod(button.getName());
		if (method == null)
			return;
		
		int left = method.getLeft();
		int top = method.getTop();
		this.method = method;
		renderables = method.getRenderables();
		rootElement = method.getRootElement();
		
		hScrollbar.setValue(0);
		vScrollbar.setValue(0);
		updateScrollbars();
		hScrollbar.setValue(left);
		vScrollbar.setValue(top);
	}

	@Override
	public void refresh() {
		repaint();
	}
	
	/**
	 * Diese Klasse ist der LocationChangeHandler fr den vertikalen
	 * Scrollbalken. Die darin enthaltene Methode onLocationChange()
	 * wird aufgerufen, sobald sich die Position des vertikalen
	 * Scrollbalkens verndert. Dabei wird der sichtbare Bereich
	 * abhngig von der Position des Scrollbakens verschoben.
	 * @author HackZ
	 *
	 */
	class VerticalLocationHandler implements LocationChangeHandler {
		@Override
		public void onLocationChanged() {
			int delta = vScrollbar.getValue() - method.getTop();
			method.setTop(vScrollbar.getValue());
			for (Renderable r : renderables)
				r.moveRelative(0, -delta);
			
			rootElement.moveRelative(0, -delta);
			
			repaint();
		}
	}
	
	/**
	 * Diese Klasse ist der LocationChangeHandler fr den horizontalen
	 * Scrollbalken. Die darin enthaltene Methode onLocationChange()
	 * wird aufgerufen, sobald sich die Position des horizontalen
	 * Scrollbalkens verndert. Dabei wird der sichtbare Bereich
	 * abhngig von der Position des Scrollbakens verschoben.
	 * @author HackZ
	 *
	 */
	class HorizontalLocationHandler implements LocationChangeHandler {
		@Override
		public void onLocationChanged() {
			int delta = hScrollbar.getValue() - method.getLeft();
			method.setLeft(hScrollbar.getValue());
			for (Renderable r : renderables)
				r.moveRelative(-delta, 0);
			
			rootElement.moveRelative(-delta, 0);
			
			repaint();
		}
	}
	
	/**
	 * Dieser ClickHandler wird an das ContextMenu bergeben und die
	 * darin enthaltene Methode onClick wird aufgerufen, wenn auf
	 * das MenuItem "ffnen" geklickt wurde und somit ein neues
	 * Tab geffnet werden soll.
	 * @author HackZ
	 *
	 */
	class OpenClickHandler implements ClickHandler {
		private String name;
		
		public OpenClickHandler(String name) {
			this.name = name;
		}

		@Override
		public void onClick() {
			openTab(name);
		}
	}
	
	/**
	 * Dieser ClickHandler wird an das ContextMenu bergeben und die
	 * darin enthaltene Methode onClick wird aufgerufen, wenn auf das
	 * MenuItem "umbenennen" geklickt wurde. Dies soll das RenameFrame
	 * ffnen, wo der andere Name eingegeben werden soll.
	 * @author HackZ
	 *
	 */
	class RenameClickHandler implements ClickHandler {
		private String name;
		
		public RenameClickHandler(String name) {
			this.name = name;
		}

		@Override
		public void onClick() {
			onRename(name);
		}
	}
	
	/**
	 * Dieser ClickHandler wird an das ContextMenu bergeben und die
	 * darin enthaltene Methode onClick wird aufgerufen, wenn auf das
	 * MenuItem "lschen" geklickt wurde. Dies soll das Element mit
	 * dem Namen lschen, wenn es nirgendwo gebraucht wurde.
	 * @author HackZ
	 *
	 */
	class DeleteClickHandler implements ClickHandler {
		private String name;
		
		public DeleteClickHandler(String name) {
			this.name = name;
		}

		@Override
		public void onClick() {
			onDelete(name);
		}
	}
	
	/**
	 * Dieser ClickHandler wird an das ContextMenu bergeben und die
	 * darin enthaltene Methode onClick wird aufgerufen, wenn auf das
	 * MenuItem "entfernen" geklickt wurde. Dies soll das Element
	 * vom Panel entfernen.
	 * @author HackZ
	 *
	 */
	class RemoveClickHandler implements ClickHandler {
		private Renderable elem;
		
		public RemoveClickHandler(Renderable elem) {
			this.elem = elem;
		}

		@Override
		public void onClick() {
			onRemove(elem);
		}
	}
	
	/**
	 * Dieser Handler sorgt dafr, dass neu geffnete Tabs im
	 * StorageController als offen markiert werden und geschlossene
	 * Tabs als geschlossen. Das enablet auch den Speichernbutton,
	 * da die offenen Tabs mitgespeichert werden.
	 * @author HackZ
	 *
	 */
	class OpenedMethodHandler implements TabOpenedHandler {
		public void tabOpened(String name, boolean opened) {
			getStorageController().setOpened(name, opened);
			modified();
		}
	}
	
	/**
	 * Dieser Handler wird an den StorageController bergeben, der die
	 * darin enthaltene Methode nextMethod aufruft, falls bei Schrittweiser
	 * Ausfhrung der Tab gender werden soll.
	 * @author HackZ
	 *
	 */
	class OnNextMethodHandler implements NextMethodHandler {
		public void nextMethod(String name) {
			tabPanel.addTab(name);
		}
	}
}
